/* Declarations for oni.cpp.
   Copyright (C) 2005 Yamamoto Takanori / Tatyana.

This is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This software is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this software; if not, write to the Free Software
Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <sys/types.h>
#include <sys/stat.h>
#include <direct.h>

size_t strlcpy(char *dst, const char *src, size_t siz);
size_t strlcat(char *dst, const char *src, size_t siz);
void PrepareDirTree(const char *s0);
void func(const char *src, const char *dst, size_t start, size_t size);

size_t strlcpy(char *dst, const char *src, size_t siz)
{
	register char *d = dst;
	register const char *s = src;
	register size_t n = siz;

	/* Copy as many bytes as will fit */
	if (n != 0 && --n != 0) {
		do {
			if ((*d++ = *s++) == 0)
				break;
		} while (--n != 0);
	}

	/* Not enough room in dst, add NUL and traverse rest of src */
	if (n == 0) {
		if (siz != 0)
			*d = '\0';		/* NUL-terminate dst */
		while (*s++)
			;
	}

	return(s - src - 1);	/* count does not include NUL */
}

size_t strlcat(char *dst, const char *src, size_t siz)
{
	register char *d = dst;
	register const char *s = src;
	register size_t n = siz;
	size_t dlen;

	/* Find the end of dst and adjust bytes left but don't go past end */
	while (n-- != 0 && *d != '\0')
		d++;
	dlen = d - dst;
	n = siz - dlen;

	if (n == 0)
		return(dlen + strlen(s));
	while (*s != '\0') {
		if (n != 1) {
			*d++ = *s;
			n--;
		}
		s++;
	}
	*d = '\0';

	return(dlen + (s - src));	/* count does not include NUL */
}

void PrepareDirTree(const char *s0)
{
	size_t len = strlen(s0)+1;
	char *s, *t;
	if((s=(char *)malloc(len+1))==NULL || (t=(char *)malloc(len+1))==NULL){
		fprintf(stderr, "not enough memory.\n");
		exit(0);
	}
	strlcpy(s, s0, len+1);
	strlcpy(t, "", len+1);

	char *p = s;
	while(strstr(p, "\\")){
		char *q = strtok(p, "\\");
		p=p+strlen(p)+1;
		strlcat(t, q, len+1);
		strlcat(t, "\\", len+1);
		if(mkdir(t)){
			/* mkdir failed */
		}
	}

	free(t);
	free(s);
}

#define H_BRAND_ONI 0x304B4150 /* "PAK0" */

struct _h1 {
	unsigned int a; /* BRAND */
	unsigned int b; /* file name : start address */
	unsigned int c; /* number of h2 */
	unsigned int d; /* number of h3 */
};

struct _h2 {
	unsigned int a; /* parent directory index */
	unsigned int b; /* number of files in this directory */
};

struct _h3 {
	unsigned int  a; /* file data : start address */
	unsigned int  b; /* file data : length */
	unsigned int  c; /* Unknown */
	unsigned int  d; /* Unknown */
};

int main(int argc, char *argv[])
{
	size_t size;
	struct _h1 h1;
	struct _h2 *h2;
	struct _h3 *h3;
	char **name;
	char **path;
	unsigned int i,j,k;

	if(argc==1){
		fprintf(stderr, "usage : %s (archive)\n", argv[0]);
		return 0;
	}

	FILE *fp = fopen(argv[1], "rb");
	if(fp == NULL){
		fprintf(stderr, "file open error.\n");
		return -1;
	}

	/* read header(phase#1) */
	if((size=fread(&h1, sizeof(struct _h1), 1, fp)) != 1){
		fprintf(stderr, "read error.\n");
		return -1;
	}

	if(h1.a != H_BRAND_ONI){
		fprintf(stderr, "invarid format.\n");
		return -1;
	}

	h2 = (_h2 *)malloc(sizeof(struct _h2)*h1.c);
	h3 = (_h3 *)malloc(sizeof(struct _h3)*h1.d);
	path = (char **)malloc(sizeof(char *)*h1.c);
	name = (char **)malloc(sizeof(char *)*(h1.d+h1.c-1));

	if(h2==NULL || h3==NULL || path==NULL || name==NULL){
		fprintf(stderr, "not enough memory.\n");
		exit(0);
	}


	for(i=0;i<h1.c;i++){
		/* read header(phase#2) */
		if((size=fread(h2+i, sizeof(struct _h2), 1, fp)) != 1){
			fprintf(stderr, "read error.\n");
			return -1;
		}
	}

	for(i=0;i<h1.d;i++){
		/* read header(phase#3) */
		if((size=fread(h3+i, sizeof(struct _h3), 1, fp)) != 1){
			fprintf(stderr, "read error.\n");
			return -1;
		}
	}

	i=0;
	for(i=0;i<h1.d+h1.c-1;i++){
		char len;
		if((size=fread(&len, sizeof(len), 1, fp)) != 1){
			fprintf(stderr, "read error.\n");
			return -1;
		}
		name[i] = (char *)malloc(len+1);
		if(name[i]==NULL){
			fprintf(stderr, "not enough memory.\n");
			exit(0);
		}
		memset(name[i], 0, len+1);
		if((size=fread(name[i], len, 1, fp)) != 1){
			fprintf(stderr, "read error.\n");
			return -1;
		}
	}

	// create the destination directories.
	j=0;
	for(i=0;i<h1.c;i++){
		char *p,*q;
		if(i==0){
			p="";
		}else{
			p=name[j++];
		}
		if(h2[i].a==0xffffffff){
			q=".";
		}else{
			q=path[h2[i].a];
		}
		path[i] = (char *)malloc(strlen(q)+1+strlen(p)+1);
		if(path[i]==NULL){
			fprintf(stderr, "not enough memory.\n");
			exit(0);
		}
		strlcpy(path[i], q, strlen(q)+1+strlen(p)+1);
		strlcat(path[i], p, strlen(q)+1+strlen(p)+1);
		strlcat(path[i], "\\", strlen(q)+1+strlen(p)+1);

		PrepareDirTree(path[i]);
	}


	//split the source file.
	fseek(fp, h1.b, SEEK_SET);

	for(i=0,j=0;j<h1.c;j++){
		for(k=i;k<h2[j].b;k++,i++){
			char *fn;
			int len_fn=strlen(path[j])+strlen(name[i+h1.c-1])+1;
			fn = (char *)malloc(len_fn);
			if(fn==NULL){
				fprintf(stderr, "not enough memory.\n");
				exit(0);
			}
			strlcpy(fn, path[j], len_fn);
			strlcat(fn, name[i+h1.c-1], len_fn);
			//fprintf(stdout, "FILENAME : %d:%s\n", i,fn);

			{
				FILE *fp_dst;
				char *buff;
				buff=(char *)malloc(h3[i].b);
				if(buff==NULL){
					fprintf(stderr, "not enough memory.\n");
					exit(0);
				}

				if((fp_dst=fopen(fn, "wb"))==NULL){
					fprintf(stderr, "file open error.\n");
					exit(0);
				}
				fprintf(stderr, "%s, (%d bytes).\n", fn,h3[i].b);
				if((size=fread(buff, 1, h3[i].b, fp)) != h3[i].b){
					fprintf(stderr, "read error.\n");
					return -1;
				}
				if(fwrite(buff, 1, h3[i].b, fp_dst) != size){
					fprintf(stderr, "file write error.\n");
					exit(0);
				}

				fclose(fp_dst);
				free(buff);
			}

			free(fn);
		}
	}


	//free

	fclose(fp);
	return 0;
}
